from __future__ import print_function
import optparse
import os.path
import numpy as np
import numpy.random as rnd # Note that this (probably) seeds itself differently for each run! Specifically, the documentation for numpy.random.RandomState states: "[if no explicit seed is given], then RandomState will try to read data from /dev/urandom (or the Windows analogue) if available or seed from the clock otherwise."
from PIL import Image

# Parse options
parser = optparse.OptionParser(usage="Usage: %prog [options] image", description="Create noisy version of a colour image.")
parser.add_option('--sigma', action="append", type=float, default=[], help="Standard deviation of noise (more than one can be specified, corresponding to the channels, if less are specified the last sigma is repeated).")
parser.add_option('--rho', type=float, default="0.0", help="Correlation factor of noise.")
parser.add_option("--out", type=str, default=None)
(options, args) = parser.parse_args()
if len(args) != 1:
    parser.error("Exactly one image required.")

inputFilename = args[0]
inputbasepath,inputextension = os.path.splitext(inputFilename)
inputbasename = os.path.basename(inputbasepath)
rho = np.clip(options.rho,0.0,1.0)

sigma = options.sigma if len(options.sigma)>0 else [32]
sigmaText = ('%g' % sigma[0]) if len(sigma)==1 else None
while len(sigma)<3:
    sigma.append(sigma[-1])
sigma = sigma[:3]
if sigmaText is None:
    sigmaText = "%s" % '_'.join(map(lambda s:'%g' % s,sigma))

if options.out is None:
    outputFilename = '%s-sigma=%s-rho=%g.png' % (inputbasepath,sigmaText,rho)
else:
    outputFilename = options.out
outputbasepath,outputextension = os.path.splitext(outputFilename)
outputbasename = os.path.basename(outputbasepath)

try:
    os.makedirs(os.path.dirname(outputFilename))
except OSError:
    # Should be okay, or otherwise we will fail later on anyway.
    # The reason to do this instead of checking whether the directory exists, is that the latter can lead to a race condition.
    pass

################################################
# Start doing some actual work

print("Generating %s" % outputbasename)

# Create the covariance matrix, such that the correlation between different channels is equal to rho.
# It is based on "Morphological Operations for Color Image Processing" (as that is apparently what has been used for "A comparative study on multivariate mathematical morphology.")
mean = np.zeros(3)
Sigma = np.array([[    sigma[0]*sigma[0],rho*sigma[0]*sigma[1],rho*sigma[0]*sigma[2]],
                  [rho*sigma[1]*sigma[0],    sigma[1]*sigma[1],rho*sigma[1]*sigma[2]],
                  [rho*sigma[2]*sigma[0],rho*sigma[2]*sigma[1],    sigma[2]*sigma[2]]])

# Load image
orgimg = np.array(Image.open(inputFilename))
if orgimg.shape[2]>3:
    orgimg = orgimg[:,:,:3] # Strip alpha channel if present.
elif orgimg.shape[2]!=3:
    parser.error("Only RGB(A) images are supported! (If an alpha channel is present it is ignored.)")
    
# Add noise
img = orgimg + rnd.multivariate_normal(mean,Sigma,orgimg.shape[:2])

# Save image
img = Image.fromarray(np.round(np.clip(img,0,255)).astype('uint8'))
img.save(outputFilename)

